@@ -334,7 +334,7 @@ impl Characteristic {
         Ok(())
     }
 
-    /// Acquire file descriptor and MTU for writing.
+    /// Acquire writer for writing with low overhead.
     ///
     /// It only works with characteristic that has
     /// the [write_without_response](CharacteristicFlags::write_without_response) flag set.
@@ -351,8 +351,12 @@ impl Characteristic {
     pub async fn write_io(&self) -> Result<CharacteristicWriter> {
         let options = PropMap::new();
         let (fd, mtu): (OwnedFd, u16) = self.call_method("AcquireWrite", (options,)).await?;
-        let stream = UnixStream::from_std(unsafe { std::os::unix::net::UnixStream::from_raw_fd(fd.into_fd()) })?;
-        Ok(CharacteristicWriter { mtu: mtu.into(), stream })
+        let stream = unsafe { std::os::unix::net::UnixStream::from_raw_fd(fd.into_fd()) };
+        stream.set_nonblocking(true)?;
+        let stream = UnixStream::from_std(stream)?;
+        // WORKAROUND: BlueZ drops data at end of packet if full MTU is used.
+        let mtu = mtu.saturating_sub(5).into();
+        Ok(CharacteristicWriter { mtu, stream })
     }
 
     /// Starts a notification or indication session from this characteristic
@@ -399,7 +403,7 @@ impl Characteristic {
             .await
     }
 
-    /// Acquire file descriptor and MTU for notify.
+    /// Acquire reader for notify with low overhead.
     ///
     /// It only works with characteristic that has
     /// the [notify](CharacteristicFlags::notify) flag set and no other client has called
@@ -423,8 +427,10 @@ impl Characteristic {
     pub async fn notify_io(&self) -> Result<CharacteristicReader> {
         let options = PropMap::new();
         let (fd, mtu): (OwnedFd, u16) = self.call_method("AcquireNotify", (options,)).await?;
-        let stream = UnixStream::from_std(unsafe { std::os::unix::net::UnixStream::from_raw_fd(fd.into_fd()) })?;
-        Ok(CharacteristicReader { mtu: mtu.into(), stream })
+        let stream = unsafe { std::os::unix::net::UnixStream::from_raw_fd(fd.into_fd()) };
+        stream.set_nonblocking(true)?;
+        let stream = UnixStream::from_std(stream)?;
+        Ok(CharacteristicReader { mtu: mtu.into(), stream, buf: Vec::new() })
     }
 
     dbus_interface!();
